import logging
from decimal import Decimal
from datetime import datetime
from .market_stream import MarketStream
from ....types.trade import Trade

log = logging.getLogger('aiotrading')

class TradeStream(MarketStream):

    
    def __init__(self, exchange, symbol):
        super().__init__(exchange, f'{symbol}@aggTrade')
        self.symbol = symbol
        self.type = 'trade'

    async def write(self, d):
        t = Trade(
            id = d['a'],
            time=datetime.fromtimestamp(d['T']/1000),
            price=Decimal(d['p']),
            volume=Decimal(d['q']),
            buy = d['m'],
        )
        return await self.queue.put(t)
    
    def __str__(self):
        return f'{self.symbol}'



import logging
from decimal import Decimal
from datetime import datetime
from .market_stream import MarketStream
from ....types.candle import Candle

log = logging.getLogger('aiotrading')

class CandleStream(MarketStream):

    def __init__(self, exchange, symbol, timeframe):
        super().__init__(exchange, f'{symbol}@kline_{timeframe}')
        self.symbol = symbol
        self.timeframe = timeframe
        self.type = 'canlde'

    async def write(self, d):
        c = Candle(
            open_time=datetime.fromtimestamp(d['k']['t']/1000),
            update_time=datetime.fromtimestamp(d['E']/1000),
            open=Decimal(d['k']['o']),
            high=Decimal(d['k']['h']),
            low=Decimal(d['k']['l']),
            close=Decimal(d['k']['c']),
            volume=Decimal(d['k']['v']),
            buy_volume = Decimal(d['k']['V']),
            trades=int(d['k']['n']),
            closed=d['k']['x'],
        )
        return await self.queue.put(c)
    
    def __str__(self):
        return f'{self.symbol}, {self.timeframe}'



import logging
import asyncio
import json
import websockets
from collections import defaultdict
from websockets.exceptions import ConnectionClosedOK
from .market_stream import MarketStream
from .candle_stream import CandleStream
from .trade_stream import TradeStream

log = logging.getLogger('aiotrading')

class Exchange:

    name = 'binance-futures'
    market_websocket_uri = 'wss://fstream.binance.com/'

    def __init__(self):
        log.info(f'creating exchange: {Exchange.name}')
        self.market_streams = defaultdict(set)
        self.market_streams_count = 0
        self.market_streams_lock = asyncio.Lock()

    def market_stream(self, name):
        return MarketStream(self, name)

    def candle_stream(self, symbol, timeframe):
        return CandleStream(self, symbol, timeframe)

    def trade_stream(self, symbol):
        return TradeStream(self, symbol)

    async def market_websocket_task(self):
        log.debug('starting market websocket task')
        while True:
            try:
                msg = await self.market_websocket.recv()
            except ConnectionClosedOK:
                break
            j = json.loads(msg)
            if 'stream' in j and 'data' in j:
                d = j['data']
                s = j['stream']
                if s in self.market_streams:
                    for stream in self.market_streams[s]:
                        await stream.write(d)
        log.debug('market websocket task stopped')
    
    async def open_market_stream(self, stream):
        async with self.market_streams_lock:
            name = stream.name
            self.market_streams[name].add(stream)
            self.market_streams_count += 1
            if self.market_streams_count == 1:
                uri = f'{Exchange.market_websocket_uri}stream?streams={name}'
                log.debug(f'connecting to market websocket stream of {Exchange.name} at {uri}')
                self.market_websocket = await websockets.connect(uri)
                asyncio.get_event_loop().create_task(self.market_websocket_task())
            elif len(self.market_streams[name]) == 1:
                log.debug(f'subscribing to market websocket stream of {Exchange.name} to {name}')
                msg = json.dumps({'method': 'SUBSCRIBE', 'params': [name], 'id': 1})
                await self.market_websocket.send(msg)

    async def close_market_stream(self, stream):
        async with self.market_streams_lock:
            name = stream.name
            if name not in self.market_streams or stream not in self.market_streams[name]:
                log.error(f'stream {name} is not open')
                return
            if self.market_streams_count == 1:
                log.debug(f'disconnecting from market websocket stream of {Exchange.name}')
                await self.market_websocket.close()
            elif len(self.market_streams[name]) == 1:
                log.debug(f'unsubscribing from market websocket stream of {Exchange.name} from {name}')
                msg = json.dumps({'method': 'UNSUBSCRIBE', 'params': [name], 'id': 1})
                await self.market_websocket.send(msg)        
            stream_set = self.market_streams[name]
            stream_set.remove(stream)
            self.market_streams_count -= 1
                



class Trade():

    def __init__(self, id, time, price, volume, buy):
        self.id = id
        self.time = time
        self.price = price
        self.volume = volume
        self.buy = buy

    def __str__(self):
        return f'price:{self.price}, volume:{self.volume}'

    def __repr__(self):
        return self.__str__()



import setuptools

with open('README.md', 'r', encoding='utf-8') as fh:
    readme = fh.read()

with open("requirements.txt", encoding="utf-8") as r:
    requirements = [i.strip() for i in r]

setuptools.setup(
    name='aiotrading',
    version='0.0.2',
    author='Hamid Fadishei',
    author_email='fadishei@yahoo.com',
    description='Async trading library based on asyncio',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/fadishei/aiotrading',
    license='GPLv3+',
    packages=setuptools.find_packages(),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Financial and Insurance Industry',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries',
    ],
    keywords="trading asyncio binance restful websockets",
    python_requires='>=3.6',
    install_requires=requirements,
)



class Candle():

    def __init__(self, open_time, update_time, open, high, low, close, volume, buy_volume, trades, closed):
        self.open_time = open_time
        self.update_time = update_time
        self.open = open
        self.high = high
        self.low = low
        self.close = close
        self.volume = volume
        self.buy_volume = buy_volume
        self.trades = trades
        self.closed = closed

    def __str__(self):
        return f'o:{self.open}, h:{self.high}, l:{self.low}, c:{self.close}, v:{self.volume}'

    def __repr__(self):
        return self.__str__()



import logging
import asyncio

log = logging.getLogger('aiotrading')

class MarketStream:

    def __init__(self, exchange, name):
        self.exchange = exchange
        self.name = name
        self.type = 'market'
        self.queue = asyncio.Queue()

    async def read(self):
        return await self.queue.get()

    async def write(self, data):
        return await self.queue.put(data)

    async def open(self):
        log.info(f'opening {self.type} stream: {self}')
        await self.exchange.open_market_stream(self)

    async def close(self):
        log.info(f'closing {self.type} stream: {self}')
        await self.exchange.close_market_stream(self)
        
    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()
        
    def __str__(self):
        return self.name

    def __repr__(self):
        return self.__str__()
